||----------------------------------------------------------------------|| 
||									||
||	pretty.m							||
||									||
||	Pretty printing program for the abstract syntax			||
||	used in denotational semantics.					||
||									||
||	(c) Simon Thompson, 1995					||
||									||
||----------------------------------------------------------------------|| 

%include "abstract_syntax"

||----------------------------------------------------------------------||
||									||
||	Printing is performed thus:					||
||									||
||	One command at most per line, control constructs are split	||
||	across lines in the usual way. Expressions are assumed to	||
||	fit one a single line. This seems to be the prctice for		||
||	Pascal, but could be changed easily enough. Definitions are	||
||	listed one per line, and are terminated by semi-colons. 	||
||									||
||	It is worth noting that because ';' is a separator for 		||
||	commands, we can only insert ';'s between commands explicitly	||
||	sequenced by 'Sequence', and not terminate the pretty		||
||	printing of each command with a ';'. This means in turn that 	||
||	we cannot terminate such a string with a newline, as we		||
||	may want to add a separating ';' at a higher level.		||
||									||
||----------------------------------------------------------------------||

||----------------------------------------------------------------------||
||									||
||	Pretty printing of commands					||
||									||
||	The main function is						||
||		pretty :: command -> [char]				||
||	but this uses the auxilliary					||
||		pre :: num -> command -> [char]				||
||	which carries the appropriate indentation info			||
||	in its first argument.						||
||									||
||----------------------------------------------------------------------||

pretty :: command -> [char]

pre :: num -> command -> [char]

pretty com = pre 0 com

pre n Skip = spaces n ++ "skip"
pre n (If_Then_Else e c1 c2) = spaces n ++ "if " ++ b_pretty e ++ "\n"
				++ spaces (n+2) ++ "then" ++ "\n"
				++ pre (n+2) c1 ++ "\n"
				++ spaces (n+2) ++ "else" ++ "\n"
				++ pre (n+2) c2 
pre n (While_Do b c) = spaces n ++ "while " ++ b_pretty b ++ " do" ++ "\n"
			++ pre (n+2) c
pre n (Repeat_Until c b) = spaces n ++ "repeat " ++ "\n"
			    ++ pre (n+2) c ++ "\n"
			    ++ "until " ++ b_pretty b
pre n (Sequence l) = spaces n ++ "begin" ++ "\n"
			++ pretty_list (pre (n+2)) ";\n" l ++ "\n"
			++ spaces n ++ "end" 
pre n (Assignment id e) = spaces n ++ i_pretty id ++ ":=" ++ e_pretty e
pre n (Par_assignment l)
	= spaces n ++ lhs ++ ":=" ++ rhs
		where
		lhs = pretty_list i_pretty "," l_list
		rhs = pretty_list e_pretty "," r_list
		(l_list,r_list) = split l
		split [] = ([],[])
		split ((a,b):c) = (a:x,b:y)
				  where
				  (x,y) = split c

pre n (Block dlist c) = spaces n ++ "begin block" ++ "\n"
				 ++ concat (map (d_pre (n+2)) dlist)
				 ++ pre (n+2) c ++ "\n"
				 ++ spaces n ++ "end block"

pre n (Call id elist) = spaces n ++ i_pretty id ++ "("
			++ pretty_list e_pretty "," elist ++ ")"

pre n (Call_ref id ilist) = spaces n ++ i_pretty id ++ "("
				++ pretty_list i_pretty "," ilist ++ ")"

pre n (Call_rec id ilist) = spaces n ++ i_pretty id ++ "("
				++ pretty_list i_pretty "," ilist ++ ")"

||----------------------------------------------------------------------||
||									||
||	Now for definitions. Again our main definition is of an		||
||	auxilliary function, d_pre which carries indentation 		||
||	information in its first argument.				||
||									||
||----------------------------------------------------------------------||

d_pretty :: def -> [char]

d_pre :: num -> def -> [char]

d_pretty = d_pre 0

d_pre n (Const id ex) = spaces n ++ "const " ++ i_pretty id
			++ " = " ++ e_pretty ex ++ " ;\n"

d_pre n (Variable id) = spaces n ++ "var " ++ i_pretty id ++ " ;\n"

d_pre n (Procedure id ilist c)
	= spaces n ++ "procedure " ++ i_pretty id ++ "(" ++
	  pretty_list i_pretty "," ilist ++ ") ;\n" ++
	  pre (n+2) c ++ ";\n"

d_pre n (Procedure_ref id ilist c)
	= spaces n ++ "ref_procedure " ++ i_pretty id ++ "(" ++
	  pretty_list i_pretty "," ilist ++ ") ;\n" ++
	  pre (n+2) c ++ ";\n"

d_pre n (Procedure_rec id ilist c)
	= spaces n ++ "rec_procedure " ++ i_pretty id ++ "(" ++
	  pretty_list i_pretty "," ilist ++ ") ;\n" ++
	  pre (n+2) c ++ ";\n"

pretty_list :: (* -> [char]) -> [char] -> [*] -> [char]

||----------------------------------------------------------------------||
||	pretty_list prettifies every item in the list (3rd arg.)	||
||	using the function f, and flattens the result, separating	||
||	items with sep.							||
||----------------------------------------------------------------------||

pretty_list f sep [] = []
pretty_list f sep [a] = f a
pretty_list f sep (a:x) = (f a) ++ sep ++ pretty_list f sep x

||----------------------------------------------------------------------||
||	Pretty printing expressions.					||
||----------------------------------------------------------------------||

b_pretty (Bool tv) = show tv
b_pretty (Apply_rel r e1 e2) = "(" ++ e_pretty e1 ++ r_pretty r 
				++ e_pretty e2 ++ ")"
b_pretty (Apply_bop e e1 e2) = "(" ++ b_pretty e1 ++ bop_pretty e 
				++ b_pretty e2 ++ ")"
b_pretty (Apply_monbop e e1) = "(" ++ monbop_pretty e 
				++ b_pretty e1 ++ ")"

r_pretty Greater = ">"
r_pretty Greater_eq = ">="
r_pretty Equal = "="

bop_pretty And = " and "
bop_pretty Or  = " or "

monbop_pretty Not = "not "

i_pretty (Name st) = st

e_pretty (Con id) = i_pretty id
e_pretty (Var id) = i_pretty id
e_pretty (Num n) = show n
e_pretty (Apply_nop op e1 e2) = "("++ e_pretty e1 ++ nop_pretty op
				++ e_pretty e2 ++ ")"
e_pretty (Apply_monop op e) =  monop_pretty op ++ "(" ++ e_pretty e ++ ")"

nop_pretty Add = "+"
nop_pretty Subtract = "-"
nop_pretty Multiply = "*"

monop_pretty Abs = "abs"
monop_pretty Minus = "-"

