||----------------------------------------------------------------------|| 
||									||
||	compile.m							||
||									||
||	(c) Simon Thompson, 1995					||
||									||
||	A compiler for the command syntax found in the file		||
||		../full/abstract_syntax					||
||	Note that the file 						||
||		../full/examples					||
||	can be re-used.							||
||									||
||----------------------------------------------------------------------|| 

%include "code"

%include "../full/abstract_syntax"
%include "../basic/semantic_types" { ident == ident
				   }
||----------------------------------------------------------------------|| 
||	First we compile expressions and boolean expressions. 		||
||	Intermediate values are held on the stack.			||
||----------------------------------------------------------------------|| 

compile_expr :: expr -> [m_code]

compile_expr (Var ide) = [ Contents ide ]
compile_expr (Num n)   = [ Push_num n ]
compile_expr (Apply_nop f e1 e2)
	= compile_expr e1 ++ compile_expr e2 ++ [ Do_nop f ]
compile_expr (Apply_monop f e)
	= compile_expr e ++ [ Do_monop f ]

compile_b_expr :: b_expr -> [m_code]

compile_b_expr (Bool b)   = [ Push_bool b ]
compile_b_expr (Apply_rel f e1 e2)
	= compile_expr e1 ++ compile_expr e2 ++ [ Do_rel f ]
compile_b_expr (Apply_bop f e1 e2)
	= compile_b_expr e1 ++ compile_b_expr e2 ++ [ Do_bop f ]
compile_b_expr (Apply_monbop f e)
	= compile_b_expr e ++ [ Do_monbop f ]

||----------------------------------------------------------------------|| 
||	compile is the top-level compilation command. It uses an	||
||	auxiliary function 						||
||		compile_command						||
||	which takes an extra argument, which yields unique (new) 	||
||	for each instance which requires them.				||
||									||
||	genlabels provides these labels, which are presented in a	||
||	tree form, mirroring the structure of the expression itself.	||
||	This tree is of type						||
||		labeltree						||
||	defined below.
||----------------------------------------------------------------------|| 

||----------------------------------------------------------------------|| 
||	In reading the definitions initially, it is simplest to 	||
||	ignore (recursively) the second (labeltree) argument,		||
||	and to read them thus:						||
||									||
||	compile_command (While_Do be c) 				||
||	= [ Label newlab ]  ++ compile_b_expr be ++ 			||
||		[ Gofalse newlab' ] ++ compile_command c 		||
||		++ [ Goto newlab , Label newlab' ]			||
||									||
||	Two new labels are needed in this compilation, 			||
||		newlab, newlab'						||
||	assuming them given, we simply build code as above. In fact	||
||	we are given them by the first two components of the tree	||
||		Whiletree newlab newlab' tree				||
||	while the `tree' component is passed in to supply the labels 	||
||	for the compilation of the command `c'.				||
||----------------------------------------------------------------------|| 

compile :: command -> [m_code]

compile c = compile_command c (genlabels c)

compile_command :: command -> labeltree -> [m_code]

compile_command Skip Nulltree = []

compile_command (If_Then_Else be c1 c2)
		(Iftree newlab newlab' tree1 tree2)
	= compile_b_expr be ++ [ Gofalse newlab ] ++
	  compile_command c1 tree1 ++ [ Goto newlab' , Label newlab ] ++
	  compile_command c2 tree2 ++ [ Label newlab' ]

compile_command (While_Do be c) (Whiletree newlab newlab' tree)
	= [ Label newlab ]  ++ compile_b_expr be ++ [ Gofalse newlab' ] ++
	  compile_command c tree ++ [ Goto newlab , Label newlab' ]

compile_command (Repeat_Until c be) (Repeatree newlab tree)
	= [ Label newlab ] ++ compile_command c tree ++
	  compile_b_expr be ++ [ Gofalse newlab ]

compile_command (Sequence comms) (Seqtree lablist)
	= concat ( map2 compile_command comms lablist )

compile_command (Assignment ide e) Nulltree
	= compile_expr e ++ [ Lval ide , Assign ]

||----------------------------------------------------------------------|| 
||	The type of labeltrees						||
||----------------------------------------------------------------------|| 

labeltree ::= Nulltree |
	      Iftree label label labeltree labeltree |
	      Whiletree label label labeltree |
	      Repeatree label labeltree |
	      Seqtree [labeltree]

||----------------------------------------------------------------------|| 
|| 	Generating the labeltrees for commands.				||
||	Note that lazy evaluation will cause these to be generated 	||
||	only as they are required.					||
||----------------------------------------------------------------------|| 

genlabels :: command -> labeltree

||----------------------------------------------------------------------|| 
||	gen_aux is the function which does all the hard work. It takes	||
||	a command and the first unused label, and returns a labeltree	||
||	which has labels built starting from that label, together with	||
||	the first label unused in the new tree.				||
||									||
||	The major difficulty is in compiling lists, but even this is	||
||	quite straightforward.						||
||----------------------------------------------------------------------|| 

gen_aux :: command -> label -> ( labeltree , label )

genlabels co = fst
		where
		(fst,snd) = gen_aux co 0

gen_aux Skip n = ( Nulltree ,  n )

gen_aux (If_Then_Else be c1 c2) n
	= ( Iftree n (n+1) t1 t2 , n2 )
	  where
	  ( t1 , n1 ) = gen_aux c1 (n+2)
	  ( t2 , n2 ) = gen_aux c2 n1

gen_aux (While_Do be c) n
	= ( Whiletree n (n+1) t1 , n1 )
	  where
	  ( t1 , n1 ) = gen_aux c (n+2)

gen_aux (Repeat_Until c be) n
	= ( Repeatree n t1 , n1 )
	  where
	  ( t1 , n1 ) = gen_aux c (n+1)

gen_aux (Sequence comms) n
	= ( Seqtree seq , n' )
	  where
	  ( seq , n' ) = gen_seq comms n 
	  gen_seq [] n = ( [] , n )
	  gen_seq (c:cs) n 
		= stick t1 ( gen_seq cs n1 )
	  	  where
		  ( t1 , n1 ) = gen_aux c n
	  stick a (x,n) = ( (a:x) , n )

gen_aux (Assignment ide e) n = ( Nulltree , n )

